@@ -8,7 +8,7 @@
 #   continue.dgp2vec
 #   continue.dgp3vec
 
-# Define Continue for S3 Objects ----------------------------------------------
+# continue S3 class -----------------------------------------------------------
 #' @title Continues MCMC sampling
 #' @description Acts on a \code{gp}, \code{gpvec}, \code{dgp2}, 
 #'     \code{dgp2vec}, \code{dgp3}, or \code{dgp3vec} object.  
@@ -44,6 +44,8 @@
 #' @param re_approx logical indicating whether to re-randomize the ordering 
 #'        and update Vecchia nearest-neighbor conditioning sets (only for fits 
 #'        with \code{vecchia = TRUE})
+#' @param ord optional ordering to be used in Vecchia re-approximation (only
+#'        for fits with \code{vecchia = TRUE} when \code{re_approx = TRUE}')
 #' @param ... N/A
 #'        
 #' @return object of the same class with the new iterations appended
@@ -57,7 +59,7 @@
 continue <- function(object, new_mcmc, verb, re_approx, ...)
   UseMethod("continue", object)
 
-# Continue One Layer ----------------------------------------------------------
+# continue.gp -----------------------------------------------------------------
 #' @rdname continue
 #' @export
 
@@ -66,28 +68,31 @@ continue.gp <- function(object, new_mcmc = 1000, verb = TRUE, ...) {
   tic <- proc.time()[[3]]
   
   # Use true nugget if it was specified
-  if (length(unique(object$g)) == 1) true_g <- object$g[1] else true_g <- NULL
+  if (length(object$g) == 1) true_g <- object$g else true_g <- NULL
   
   # Start MCMC at last samples
-  initial <- list(theta = object$theta[object$nmcmc],
-                  g = object$g[object$nmcmc],
-                  tau2 = object$tau2[object$nmcmc])
+  initial <- list(theta = ifel(object$settings$sep, object$theta[object$nmcmc, ], 
+                               object$theta[object$nmcmc]),
+                  g = ifel(is.null(true_g), object$g[object$nmcmc], NULL))
   
   # Run continuing MCMC iterations
-  samples <- gibbs_one_layer(x = object$x, 
-                             y = object$y, 
-                             nmcmc = new_mcmc + 1, 
-                             verb = verb, 
-                             initial = initial, 
-                             true_g = true_g, 
-                             settings = object$settings, 
+  samples <- gibbs_one_layer(x = object$x,
+                             y = object$y,
+                             dydx = object$dydx,
+                             nmcmc = new_mcmc + 1,
+                             verb = verb,
+                             initial = initial,
+                             true_g = true_g,
+                             settings = object$settings,
                              v = object$v)
 
   # Append new information to original fit
   object$nmcmc <- object$nmcmc + new_mcmc
-  object$g <- c(object$g, samples$g[-1])
-  object$theta <- c(object$theta, samples$theta[-1])
   object$tau2 <- c(object$tau2, samples$tau2[-1])
+  if (object$settings$sep) {
+    object$theta <- rbind(object$theta, samples$theta[-1, , drop = FALSE])
+  } else object$theta <- c(object$theta, samples$theta[-1])
+  if (is.null(true_g)) object$g <- c(object$g, samples$g[-1])
   object$ll <- c(object$ll, samples$ll[-1])
   
   toc <- proc.time()[[3]]
@@ -95,46 +100,55 @@ continue.gp <- function(object, new_mcmc = 1000, verb = TRUE, ...) {
   return(object)
 }
 
-# Continue Two Layer ----------------------------------------------------------
+# continue.dgp2 ---------------------------------------------------------------
 #' @rdname continue
 #' @export
 
 continue.dgp2 <- function(object, new_mcmc = 1000, verb = TRUE, ...) {
   
   tic <- proc.time()[[3]]
+  grad_enhance <- (!is.null(object$dydx))
 
-  monowarp <- (!is.null(object$x_grid))
-  
   # Use true nugget if it was specified
-  if (length(unique(object$g)) == 1) true_g <- object$g[1] else true_g <- NULL
+  if (length(object$g) == 1) true_g <- object$g else true_g <- NULL
   
-  if (monowarp) {
-    initial <- list(w = monowarp_ref(object$x, object$x_grid, object$w_grid[[object$nmcmc]]),
-                    theta_y = object$theta_y[object$nmcmc, ],
-                    theta_w = object$theta_w[object$nmcmc, ],
-                    g = object$g[object$nmcmc],
-                    tau2 = object$tau2[object$nmcmc])
+  # Start MCMC at last samples
+  initial <- list(theta_y = object$theta_y[object$nmcmc],
+                  theta_w = object$theta_w[object$nmcmc, ],
+                  g = ifel(is.null(true_g), object$g[object$nmcmc], NULL))
+  if (object$settings$monowarp) {
+    initial$w <- NULL
+  } else {
+    initial$w <- object$w[object$nmcmc, , ]
+  }
+  if (grad_enhance) initial$dwdx <- object$dwdx
+
+
+  if (object$settings$monowarp) {
     samples <- gibbs_two_layer_mono(x = object$x,
                                     y = object$y,
                                     x_grid = object$x_grid,
                                     nmcmc = new_mcmc + 1,
-                                    D = ncol(object$theta_y),
+                                    verb = verb,
+                                    initial = initial,
+                                    true_g = true_g,
+                                    settings = object$settings,
+                                    v = object$v)
+  } else if (grad_enhance) {
+    samples <- gibbs_two_layer_grad(x = object$x,
+                                    y = object$y,
+                                    dydx = object$dydx,
+                                    nmcmc = new_mcmc + 1,
                                     verb = verb,
                                     initial = initial,
                                     true_g = true_g,
                                     settings = object$settings,
                                     v = object$v)
   } else {
-    initial <- list(w = object$w[[object$nmcmc]],
-                    theta_y = object$theta_y[object$nmcmc],
-                    theta_w = object$theta_w[object$nmcmc, ],
-                    g = object$g[object$nmcmc],
-                    tau2 = object$tau2[object$nmcmc])
-    # Run continuing MCMC iterations
     samples <- gibbs_two_layer(x = object$x, 
                                y = object$y, 
                                nmcmc = new_mcmc + 1, 
-                               D = ncol(object$w[[1]]), 
+                               D = dim(object$w)[3], 
                                verb = verb, 
                                initial = initial, 
                                true_g = true_g, 
@@ -142,28 +156,27 @@ continue.dgp2 <- function(object, new_mcmc = 1000, verb = TRUE, ...) {
                                v = object$v)
   }
   
-  
-
   # Append new information to original fit
   object$nmcmc <- object$nmcmc + new_mcmc
-  object$g <- c(object$g, samples$g[-1])
-  if (monowarp) {
-    object$theta_y <- rbind(object$theta_y, samples$theta_y[-1, , drop = FALSE])
-    object$w_grid <- c(object$w_grid, samples$w_grid[-1])
-  } else {
-    object$theta_y <- c(object$theta_y, samples$theta_y[-1])
-    object$w <- c(object$w, samples$w[-1])
+  object$tau2_y <- c(object$tau2_y, samples$tau2_y[-1])
+  object$theta_y <- c(object$theta_y, samples$theta_y[-1])
+  if (object$settings$monowarp) {
+    if (!is.null(samples$tau2_w)) {
+      object$tau2_w <- rbind(object$tau2_w, samples$tau2_w[-1, , drop = FALSE])
+    }
+    object$w_grid <- abind::abind(object$w_grid, samples$w_grid[-1, , , drop = FALSE], along = 1)
   }
   object$theta_w <- rbind(object$theta_w, samples$theta_w[-1, , drop = FALSE])
-  object$tau2 <- c(object$tau2, samples$tau2[-1])
+  object$w <- abind::abind(object$w, samples$w[-1, , , drop = FALSE], along = 1)
+  if (is.null(true_g)) object$g <- c(object$g, samples$g[-1])
   object$ll <- c(object$ll, samples$ll[-1])
-  
+
   toc <- proc.time()[[3]]
   object$time <- object$time + unname(toc - tic)
   return(object)
 }
 
-# Continue Three Layer --------------------------------------------------------
+# continue.dgp3 ---------------------------------------------------------------
 #' @rdname continue
 #' @export
 
@@ -172,22 +185,21 @@ continue.dgp3 <- function(object, new_mcmc = 1000, verb = TRUE, ...) {
   tic <- proc.time()[[3]]
   
   # Use true nugget if it was specified
-  if (length(unique(object$g)) == 1) true_g <- object$g[1] else true_g <- NULL
+  if (length(object$g) == 1) true_g <- object$g else true_g <- NULL
   
   # Start MCMC at last samples
-  initial <- list(w = object$w[[object$nmcmc]],
-                  z = object$z[[object$nmcmc]],
+  initial <- list(w = object$w[object$nmcmc, , ],
+                  z = object$z[object$nmcmc, , ],
                   theta_y = object$theta_y[object$nmcmc],
                   theta_w = object$theta_w[object$nmcmc, ],
                   theta_z = object$theta_z[object$nmcmc, ],
-                  g = object$g[object$nmcmc],
-                  tau2 = object$tau2[object$nmcmc])
+                  g = ifel(is.null(true_g), object$g[object$nmcmc], NULL))
   
   # Run continuing MCMC iterations
   samples <- gibbs_three_layer(x = object$x, 
                                y = object$y, 
                                nmcmc = new_mcmc + 1, 
-                               D = ncol(object$w[[1]]), 
+                               D = dim(object$w)[3], 
                                verb = verb, 
                                initial = initial, 
                                true_g = true_g, 
@@ -196,13 +208,13 @@ continue.dgp3 <- function(object, new_mcmc = 1000, verb = TRUE, ...) {
   
   # Append new information to original fit
   object$nmcmc <- object$nmcmc + new_mcmc
-  object$g <- c(object$g, samples$g[-1])
+  object$tau2_y <- c(object$tau2_y, samples$tau2_y[-1])
   object$theta_y <- c(object$theta_y, samples$theta_y[-1])
   object$theta_w <- rbind(object$theta_w, samples$theta_w[-1, , drop = FALSE])
   object$theta_z <- rbind(object$theta_z, samples$theta_z[-1, , drop = FALSE])
-  object$w <- c(object$w, samples$w[-1])
-  object$z <- c(object$z, samples$z[-1])
-  object$tau2 <- c(object$tau2, samples$tau2[-1])
+  object$w <- abind::abind(object$w, samples$w[-1, , , drop = FALSE], along = 1)
+  object$z <- abind::abind(object$z, samples$z[-1, , , drop = FALSE], along = 1)
+  if (is.null(true_g)) object$g <- c(object$g, samples$g[-1])
   object$ll <- c(object$ll, samples$ll[-1])
   
   toc <- proc.time()[[3]]
@@ -210,46 +222,45 @@ continue.dgp3 <- function(object, new_mcmc = 1000, verb = TRUE, ...) {
   return(object)
 }
 
-# Continue One Layer Vecchia --------------------------------------------------
+# continue.gpvec --------------------------------------------------------------
 #' @rdname continue
 #' @export
 
 continue.gpvec <- function(object, new_mcmc = 1000, verb = TRUE, 
-                           re_approx = FALSE, ...) {
+                           re_approx = FALSE, ord = NULL, ...) {
   
   tic <- proc.time()[[3]]
-  
+
   # Use true nugget if it was specified
-  if (length(unique(object$g)) == 1) true_g <- object$g[1] else true_g <- NULL
+  if (length(object$g) == 1) true_g <- object$g else true_g <- NULL
   
   # Start MCMC at last samples
-  initial <- list(theta = object$theta[object$nmcmc],
-                  g = object$g[object$nmcmc],
-                  tau2 = object$tau2[object$nmcmc])
-  
-  # Optionally redo approximation
-  if (re_approx) {
-    x_approx <- NULL
-  } else x_approx <- object$x_approx
+  initial <- list(theta = ifel(object$settings$sep, object$theta[object$nmcmc, ], 
+                               object$theta[object$nmcmc]),
+                  g = ifel(is.null(true_g), object$g[object$nmcmc], NULL))
   
   # Run continuing MCMC iterations
   samples <- gibbs_one_layer_vec(x = object$x, 
                                  y = object$y, 
+                                 dydx = object$dydx,
                                  nmcmc = new_mcmc + 1, 
                                  verb = verb, 
                                  initial = initial, 
                                  true_g = true_g, 
                                  settings = object$settings, 
                                  v = object$v, 
-                                 m = object$m, 
-                                 ordering = object$ordering,
-                                 x_approx = x_approx)
+                                 m = object$x_approx$m, 
+                                 ord = ifel(re_approx, ord, NULL), # only used if re_approx = TRUE
+                                 cores = object$x_approx$cores,
+                                 x_approx = ifel(re_approx, NULL, object$x_approx))
   
   # Append new information to original fit
   object$nmcmc <- object$nmcmc + new_mcmc
-  object$g <- c(object$g, samples$g[-1])
-  object$theta <- c(object$theta, samples$theta[-1])
   object$tau2 <- c(object$tau2, samples$tau2[-1])
+  if (object$settings$sep) {
+    object$theta <- rbind(object$theta, samples$theta[-1, , drop = FALSE])
+  } else object$theta <- c(object$theta, samples$theta[-1])
+  if (is.null(true_g)) object$g <- c(object$g, samples$g[-1])
   object$x_approx <- samples$x_approx
   object$ll <- c(object$ll, samples$ll[-1])
   
@@ -258,85 +269,90 @@ continue.gpvec <- function(object, new_mcmc = 1000, verb = TRUE,
   return(object)
 }
 
-# Continue Two Layer Vecchia --------------------------------------------------
+# continue.dgp2vec ------------------------------------------------------------
 #' @rdname continue
 #' @export
 
 continue.dgp2vec <- function(object, new_mcmc = 1000, verb = TRUE, 
-                             re_approx = FALSE, ...) {
+                             re_approx = FALSE, ord = NULL, ...) {
   
   tic <- proc.time()[[3]]
+  grad_enhance <- (!is.null(object$dydx))
 
-  monowarp <- (!is.null(object$x_grid))
-  
   # Use true nugget if it was specified
-  if (length(unique(object$g)) == 1) true_g <- object$g[1] else true_g <- NULL
-  
-  # Optionally redo approximation
-  if (re_approx) {
-    if (!monowarp) x_approx <- NULL
-    w_approx <- NULL
-  } else {
-    if (!monowarp) x_approx <- object$x_approx
-    w_approx <- object$w_approx
-  }
+  if (length(object$g) == 1) true_g <- object$g else true_g <- NULL
 
   # Start MCMC at last samples
-  if (monowarp) {
-    inital <- list(w = monowarp_ref(object$x, object$xgrid, object$w_grid[[object$nmcmc]]),
-                   theta_y = object$theta_y[object$nmcmc, ],
-                   theta_w = object$theta_w[object$nmcmc, ],
-                   g = object$g[object$nmcmc],
-                   tau2 = object$tau2[object$nmcmc])
-
+  initial <- list(theta_y = object$theta_y[object$nmcmc],
+                  theta_w = object$theta_w[object$nmcmc, ],
+                  g = ifel(is.null(true_g), object$g[object$nmcmc], NULL))
+  if (object$settings$monowarp) {
+    initial$w <- NULL
+  } else {
+    initial$w <- object$w[object$nmcmc, , ]
+  }
+  if (grad_enhance) initial$dwdx <- object$dwdx
+  
+  if (object$settings$monowarp) {
     samples <- gibbs_two_layer_vec_mono(x = object$x,
                                         y = object$y,
                                         x_grid = object$x_grid,
                                         nmcmc = new_mcmc + 1,
-                                        D = ncol(object$theta_y),
                                         verb = verb,
                                         initial = initial,
                                         true_g = true_g,
                                         settings = object$settings,
                                         v = object$v,
-                                        m = object$m,
-                                        ordering = object$ordering,
-                                        w_approx = w_approx)
+                                        m = object$w_approx$m,
+                                        ord = ifel(re_approx, ord, NULL), # only used if re_approx = TRUE
+                                        cores = object$w_approx$cores,
+                                        w_approx = ifel(re_approx, NULL, object$w_approx))
+  } else if (grad_enhance) {
+    samples <- gibbs_two_layer_vec_grad(x = object$x,
+                                        y = object$y,
+                                        dydx = object$dydx,
+                                        nmcmc = new_mcmc + 1,
+                                        verb = verb,
+                                        initial = initial,
+                                        true_g = true_g,
+                                        settings = object$settings,
+                                        v = object$v,
+                                        m = object$w_approx$m,
+                                        ord = ifel(re_approx, ord, NULL), # only used if re_approx = TRUE
+                                        cores = object$w_approx$cores,
+                                        x_approx = ifel(re_approx, NULL, object$x_approx),
+                                        w_approx = ifel(re_approx, NULL, object$w_approx))
   } else {
-    initial <- list(w = object$w[[object$nmcmc]],
-                    theta_y = object$theta_y[object$nmcmc],
-                    theta_w = object$theta_w[object$nmcmc, ],
-                    g = object$g[object$nmcmc],
-                    tau2 = object$tau2[object$nmcmc])
-
     samples <- gibbs_two_layer_vec(x = object$x, 
                                    y = object$y, 
                                    nmcmc = new_mcmc + 1, 
-                                   D = ncol(object$w[[1]]), 
+                                   D = dim(object$w)[3], 
                                    verb = verb, 
                                    initial = initial, 
                                    true_g = true_g, 
                                    settings = object$settings, 
                                    v = object$v, 
-                                   m = object$m, 
-                                   ordering = object$ordering,
-                                   x_approx = x_approx, 
-                                   w_approx = w_approx)
+                                   m = object$x_approx$m, 
+                                   ord = ifel(re_approx, ord, NULL), # only used if re_approx = TRUE
+                                   cores = object$x_approx$cores,
+                                   x_approx = ifel(re_approx, NULL, object$x_approx), 
+                                   w_approx = ifel(re_approx, NULL, object$w_approx))
   }
   
   # Append new information to original fit
   object$nmcmc <- object$nmcmc + new_mcmc
-  object$g <- c(object$g, samples$g[-1])
-  if (monowarp) {
-    object$theta_y <- rbind(object$theta_y, samples$theta_y[-1, , drop = FALSE])
-    object$w_grid <- c(object$w_grid, samples$w_grid[-1])
-  } else {
-    object$theta_y <- c(object$theta_y, samples$theta_y[-1])
-    object$w <- c(object$w, samples$w[-1])
-  }
+  object$tau2_y <- c(object$tau2_y, samples$tau2[-1])
+  object$theta_y <- c(object$theta_y, samples$theta_y[-1])
   object$theta_w <- rbind(object$theta_w, samples$theta_w[-1, , drop = FALSE])
-  object$tau2 <- c(object$tau2, samples$tau2[-1])
-  if (monowarp) object$x_approx <- samples$x_approx
+  if (object$settings$monowarp) {
+    if (!is.null(samples$tau2_w)) {
+      object$tau2_w <- rbind(object$tau2_w, samples$tau2_w[-1, , drop = FALSE])
+    }
+    object$w_grid <- abind::abind(object$w_grid, samples$w_grid[-1, , , drop = FALSE], along = 1)
+  }
+  object$w <- abind::abind(object$w, samples$w[-1, , , drop = FALSE], along = 1)
+  if (is.null(true_g)) object$g <- c(object$g, samples$g[-1])
+  if (!object$settings$monowarp) object$x_approx <- samples$x_approx
   object$w_approx <- samples$w_approx
   object$ll <- c(object$ll, samples$ll[-1])
   
@@ -345,63 +361,52 @@ continue.dgp2vec <- function(object, new_mcmc = 1000, verb = TRUE,
   return(object)
 }
 
-# Continue Three Layer Vecchia ------------------------------------------------
+# continue.dgp3vec ------------------------------------------------------------
 #' @rdname continue
 #' @export
 
 continue.dgp3vec <- function(object, new_mcmc = 1000, verb = TRUE, 
-                             re_approx = FALSE, ...) {
+                             re_approx = FALSE, ord = NULL, ...) {
   
   tic <- proc.time()[[3]]
   
   # Use true nugget if it was specified
-  if (length(unique(object$g)) == 1) true_g <- object$g[1] else true_g <- NULL
+  if (length(object$g) == 1) true_g <- object$g else true_g <- NULL
   
   # Start MCMC at last samples
-  initial <- list(w = object$w[[object$nmcmc]],
-                  z = object$z[[object$nmcmc]],
+  initial <- list(w = object$w[object$nmcmc, , ],
+                  z = object$z[object$nmcmc, , ],
                   theta_y = object$theta_y[object$nmcmc],
                   theta_w = object$theta_w[object$nmcmc, ],
                   theta_z = object$theta_z[object$nmcmc, ],
-                  g = object$g[object$nmcmc],
-                  tau2 = object$tau2[object$nmcmc])
-  
-  # Optionally redo approximation
-  if (re_approx) {
-    x_approx <- NULL
-    z_approx <- NULL
-    w_approx <- NULL
-  } else {
-    x_approx <- object$x_approx
-    z_approx <- object$z_approx
-    w_approx <- object$w_approx
-  }
+                  g = ifel(is.null(true_g), object$g[object$nmcmc], NULL))
   
   # Run continuing MCMC iterations
   samples <- gibbs_three_layer_vec(x = object$x, 
                                    y = object$y, 
                                    nmcmc = new_mcmc + 1, 
-                                   D = ncol(object$w[[1]]),
+                                   D = dim(object$w)[3],
                                    verb = verb, 
                                    initial = initial, 
                                    true_g = true_g, 
                                    settings = object$settings, 
                                    v = object$v, 
-                                   m = object$m, 
-                                   ordering = object$ordering,
-                                   x_approx = x_approx, 
-                                   z_approx = z_approx, 
-                                   w_approx = w_approx)
+                                   m = object$x_approx$m, 
+                                   ord = ifel(re_approx, ord, NULL), # only used if re_approx = TRUE
+                                   cores = object$x_approx$cores,
+                                   x_approx = ifel(re_approx, NULL, object$x_approx), 
+                                   z_approx = ifel(re_approx, NULL, object$z_approx), 
+                                   w_approx = ifel(re_approx, NULL, object$w_approx))
   
   # Append new information to original fit
   object$nmcmc <- object$nmcmc + new_mcmc
-  object$g <- c(object$g, samples$g[-1])
+  object$tau2_y <- c(object$tau2_y, samples$tau2_y[-1])
   object$theta_y <- c(object$theta_y, samples$theta_y[-1])
   object$theta_w <- rbind(object$theta_w, samples$theta_w[-1, , drop = FALSE])
   object$theta_z <- rbind(object$theta_z, samples$theta_z[-1, , drop = FALSE])
-  object$w <- c(object$w, samples$w[-1])
-  object$z <- c(object$z, samples$z[-1])
-  object$tau2 <- c(object$tau2, samples$tau2[-1])
+  if (is.null(true_g)) object$g <- c(object$g, samples$g[-1])
+  object$w <- abind::abind(object$w, samples$w[-1, , , drop = FALSE], along = 1)
+  object$z <- abind::abind(object$z, samples$z[-1, , , drop = FALSE], along = 1)
   object$x_approx <- samples$x_approx
   object$z_approx <- samples$z_approx
   object$w_approx <- samples$w_approx
